import {Endpoint, EndpointOperations} from './endpoint';
import {BaseOperationParams, Operation} from './operation';

export interface BaseAppState {
	endpoints: unknown;
	operations: unknown;
}

export type AppOperations<T extends App> = (
	T extends App<any, infer R>
		? R extends BaseAppState
			? keyof R['operations']
			: string
		: string
	);

export interface App<AppName extends string = string, AppState extends BaseAppState = {
	endpoints: [];
	operations: Record<string, []>;
}> {
	name: AppName;
	operations: Set<Operation>;
	endpoints: Set<Endpoint>;
	operation<
		OperationName extends string,
		OperationParams extends BaseOperationParams<OperationName>,
		NewOperation extends Operation<OperationParams>
	>(newOperation: NewOperation): App<
		AppName,
		{
			endpoints: AppState['endpoints'],
			operations: keyof AppState['operations'] extends never ? {
				[Key in NewOperation['name']]: (
					Exclude<NewOperation['args'], undefined> extends readonly string[] ? Exclude<NewOperation['args'], undefined> : never[]
					)
			} : {
				[Key in NewOperation['name'] | keyof AppState['operations']]: (
					AppState['operations'] extends Record<Key, any>
						? (
							AppState['operations'][Key] extends readonly string[] ? AppState['operations'][Key] : (
								Exclude<NewOperation['args'], undefined> extends readonly string[] ? Exclude<NewOperation['args'], undefined> : never[]
								)
							)
						: (
							Exclude<NewOperation['args'], undefined> extends readonly string[] ? Exclude<NewOperation['args'], undefined> : never[]
							)
					);
			}
		}
	>;
	endpoint<NewEndpoint extends Endpoint = Endpoint>(newEndpoint: EndpointOperations<NewEndpoint> extends AppOperations<this> ? NewEndpoint : never): App<
		AppName,
		{
			endpoints: AppState['endpoints'] extends Array<unknown> ? [...AppState['endpoints'], NewEndpoint] : [NewEndpoint],
			operations: AppState['operations']
		}
	>;
}

interface AppParams<Name extends string = string> {
	name: Name;
}

class AppInstance<Params extends AppParams, State extends BaseAppState> implements App<Params['name'], State> {
	readonly name: Params['name'];
	readonly endpoints: Set<Endpoint>;
	readonly operations: Set<Operation>;

	constructor(params: Params) {
		this.name = params.name;
		this.endpoints = new Set<Endpoint>();
		this.operations = new Set<Operation>();
	}

	operation<NewOperation extends Operation>(newOperation: NewOperation) {
		this.operations.add(newOperation);

		return this;
	}

	endpoint<NewEndpoint extends Endpoint = Endpoint>(newEndpoint: NewEndpoint) {
		this.endpoints.add(newEndpoint);

		return this;
	}
}

export const app = <Params extends AppParams>(params: Params): App<Params['name']> => {
	return new AppInstance(params);
};
