export const AVAILABLE_METHODS = [
	'HEAD',
	'GET',
	'POST',
	'PUT',
	'PATCH',
	'DELETE',
	'OPTIONS'
] as const;

export const AVAILABLE_EXTENSION_METHODS = [
	'QUERY'
] as const;

export type Method = typeof AVAILABLE_METHODS[number];

export type MethodWithExtensions = Method | typeof AVAILABLE_EXTENSION_METHODS[number];

export interface BaseOperationParams<
	Name extends string = string,
	Method extends MethodWithExtensions = MethodWithExtensions,
	Args extends readonly string[] = readonly string[]
> {
	name: Name;
	method?: Method;
	args?: Args;
}

export interface Operation<Params extends BaseOperationParams = BaseOperationParams> {
	name: Params['name'];
	method: Params['method'];
	args: Params['args'];
}

class OperationInstance<Params extends BaseOperationParams = BaseOperationParams> implements Operation<Params> {
	readonly name: Params['name'];
	readonly method: Params['method'];
	readonly args: Params['args'];

	constructor(params: Params) {
		this.name = params.name;
		this.method = params.method ?? 'GET';
		this.args = params.args;
	}
}

export const operation = <Params extends BaseOperationParams = BaseOperationParams>(
	params: Params
): Operation<Params> => {
	return new OperationInstance(params);
};
